//! Implementation of manipulation of Advanced Technology Attachment (ATA) compliant devices

use alloc::sync::Arc;
use alloc::{format, vec};
use core::mem;

use anyhow::{Ok, Result};
use bitflags::bitflags;
use log::{error, info, trace, warn};
use spin::Mutex;
use x86_64::instructions::port::{Port, PortReadOnly, PortWriteOnly};

use super::block::Manipulation;
use super::pci;
use super::storage::{Controller, Device};
use crate::kernel::device::pci::Header;

/// Size in bytes of the [`IdentifyDeviceData`] structure, defined in [Microsoft documentation](https://learn.microsoft.com/en-us/windows-hardware/drivers/ddi/ata/ns-ata-_identify_device_data)
const IDENTIFY_DATA_INFORMATIONS_SIZE: usize = 512;

/// Size of a block in bytes
pub const BLOCK_SIZE: usize = 512;

/// Maximal value of LBA if encoded on 28 bits
const MAX_LBA_28_VALUE: usize = (1 << 28) - 1;

/// Default port number for the primary ATA bus
const DEFAULT_PRIMARY_BUS_REGISTER: u16 = 0x01F0;

/// Default port number for the primary ATA bus
const DEFAULT_SECONDARY_BUS_REGISTER: u16 = 0x03F6;

/// Default port number for the primary ATA bus
const DEFAULT_PRIMARY_CONTROL_STATUS_REGISTER: u16 = 0x01F0;

/// Default port number for the primary ATA bus
const DEFAULT_SECONDARY_CONTROL_STATUS_REGISTER: u16 = 0x376;

/// Commands that can be sent to an ATA drive's command port
///
/// See the [OSdev wiki](https://wiki.osdev.org/ATA_Command_Matrix) for more informations
#[repr(u8)]
#[derive(Debug)]
#[allow(unused, missing_docs)]
enum Command {
    Nop = 0x00,

    CfaRequestExtendedErrorCode = 0x03,

    DataSetManagement = 0x06,

    DataSetManagementXL = 0x07,

    DeviceReset = 0x08,

    RequestSenseDataExt = 0x0B,

    Recalibrate10 = 0x10,

    Recalibrate11 = 0x11,

    GetPhysicalElementStatus = 0x12,

    Recalibrate13 = 0x13,

    Recalibrate14 = 0x14,

    Recalibrate15 = 0x15,

    Recalibrate16 = 0x16,

    Recalibrate17 = 0x17,

    Recalibrate18 = 0x18,

    Recalibrate19 = 0x19,

    Recalibrate1A = 0x1A,

    Recalibrate1B = 0x1B,

    Recalibrate1C = 0x1C,

    Recalibrate1D = 0x1D,

    Recalibrate1E = 0x1E,

    Recalibrate1F = 0x1F,

    /// Reads sectors using PIO (28-bit LBA)
    ReadSectors = 0x20,

    ReadSectorsWithoutRetry = 0x21,

    ReadLong = 0x22,

    ReadLongWithoutRetry = 0x23,

    /// Reads sectors using PIO (48-bit LBA)
    ReadSectorsExt = 0x24,

    /// Reads sectors using DMA (48-bit LBA)
    ReadDmaExt = 0x25,

    ReadDmaQueuedExt = 0x26,

    ReadNativeMaxAddressExt = 0x27,

    ReadMultipleExt = 0x29,

    ReadStreamDmaExt = 0x2A,

    ReadLogExt = 0x2F,

    /// Writes sectors using PIO (28-bit LBA)
    WriteSectors = 0x30,

    WriteSectorsWithoutRetry = 0x31,

    WriteLong = 0x32,

    WriteLongWithoutRetry = 0x33,

    /// Writes sectors using PIO (48-bit LBA)
    WriteSectorsExt = 0x34,

    /// Writes sectors using DMA (48-bit LBA)
    WriteDmaExt = 0x35,

    WriteDmaQueuedExt = 0x36,

    SetMaxAddressExt = 0x37,

    CFAWriteSectorsWithoutErase = 0x38,

    WriteMultipleExt = 0x39,

    WriteStreamDmaExt = 0x3A,

    WriteStreamExt = 0x3B,

    WriteVerify = 0x3C,

    WriteDmaFuaExt = 0x3D,

    WriteDmaQueuedFuaExt = 0x3E,

    WriteLogExt = 0x3F,

    ReadVerifySectors = 0x40,

    ReadVerifySectorsWithoutRetry = 0x41,

    ReadVerifySectorsExt = 0x42,

    ZeroExt = 0x44,

    WriteUncorrectableExt = 0x45,

    ReadLogDmaExt = 0x47,

    ZacManagementIn = 0x4A,

    FormatTrack = 0x50,

    ConfigureStream = 0x51,

    WriteLogDmaExt = 0x57,

    TrustedNonData = 0x5B,

    TrustedReceive = 0x5C,

    TrustedReceiveDma = 0x5D,

    TrustedSend = 0x5E,

    TrustedSendDma = 0x5F,

    ReadFPDmaQueued = 0x60,

    WriteFPDmaQueued = 0x61,

    NCQNonData = 0x63,

    SendFPDmaQueued = 0x64,

    REceiveFPDmaQueued = 0x65,

    Seek70 = 0x70,

    Seek71 = 0x71,

    Seek72 = 0x72,

    Seek73 = 0x73,

    Seek74 = 0x74,

    Seek75 = 0x75,

    Seek76 = 0x76,

    Seek77AndSetDateAndTimeExt = 0x77,

    Seek78AndAccessibleMaxAddressConfiguration = 0x78,

    Seek79 = 0x79,

    Seek7A = 0x7A,

    Seek7B = 0x7B,

    Seek7CAndRemoveElementAndTruncate = 0x7C,

    Seek7DAndRestoreElementAndRebuild = 0x7D,

    Seek7EAndRemoveElementAndModifyZones = 0x7E,

    Seek7F = 0x7F,

    CfaTranslateSector = 0x87,

    ExecuteDeviceDiagnostic = 0x90,

    InitializeDeviceParameters = 0x91,

    DownloadMicrocode = 0x92,

    DownloadMicrocodeDma = 0x93,

    StandbyImmediate94 = 0x94,

    IdleImmediate95 = 0x95,

    MutateExtAndStandby = 0x96,

    Idle97 = 0x97,

    CheckPowerMode98 = 0x98,

    Sleep99 = 0x99,

    ZacManagementOut = 0x9F,

    /// Sends a packet, for ATAPI devices using the packet interface (PI)
    Packet = 0xA0,

    /// Gets identifying details of an ATAPI drive
    IdentifyPacketDevice = 0xA1,

    Service = 0xA2,

    Smart = 0xB0,

    DeviceConfigurationOverlay = 0xB1,

    SetSectorConfigurationExt = 0xB2,

    SanitizeDevice = 0xB4,

    NVCache = 0xB6,

    CfaEraseSectors = 0xC0,

    ReadMultiple = 0xC4,

    WriteMultiple = 0xC5,

    SetMultipleMode = 0xC6,

    ReadDmaQueued = 0xC7,

    /// Reads sectors using DMA (28-bit LBA)
    ReadDma = 0xC8,

    ReadDmaWithoutRetry = 0xC9,

    /// Writes sectors using DMA (28-bit LBA)
    WriteDma = 0xCA,

    WriteDmaWithoutRetry = 0xCB,

    WriteDmaQueued = 0xCC,

    CfaWriteMultipleWithoutErase = 0xCD,

    WriteMultipleFUAExt = 0xCE,

    CheckMediaCardType = 0xD1,

    GetMediaStatus = 0xDA,

    AcknowledgeMediaChange = 0xDB,

    BootPostBoot = 0xDC,

    BootPreBoot = 0xDD,

    MediaLock = 0xDE,

    StandbyImmediateE0 = 0xE0,

    IdleImmediateE1 = 0xE1,

    Standby = 0xE2,

    IdleE3 = 0xE3,

    ReadBuffer = 0xE4,

    CheckPowerModeE5 = 0xE5,

    SleepE6 = 0xE6,

    /// Flushes the drive's bus cache (28-bit LBA)
    ///
    /// It is needed to use this command after each write.
    FlushCache = 0xE7,

    WriteBuffer = 0xE8,

    ReadBufferDmaAndWriteSame = 0xE9,

    /// Flushes the drive's bus cache (48-bit LBA)
    ///
    /// It is needed to use this command after each write.
    FlushCacheExt = 0xEA,

    WriteBufferDma = 0xEB,

    /// Gets identifying details of an ATA drive
    IdentifyDevice = 0xEC,

    MediaEject = 0xED,

    IdentifyDeviceDma = 0xEE,

    SetFeatures = 0xEF,

    SecuritySetPassword = 0xF1,

    SecurityUnlock = 0xF2,

    SecurityErasePrepare = 0xF3,

    SecurityEraseUnit = 0xF4,

    SecurityFreezeLock = 0xF5,

    SecurityDisablePassword = 0xF6,

    ReadNativeMaxAddress = 0xF8,

    SetMaxAddress = 0xF9,
}

impl From<Command> for u8 {
    fn from(value: Command) -> Self {
        value as Self
    }
}

/// Kind of drives attached to a bus
#[repr(u8)]
#[derive(Debug, Clone, Copy)]
enum DriveRole {
    /// Master drive
    Master = 0x00,

    /// Slave drive
    Slave = 0x10,
}

impl From<DriveRole> for u8 {
    fn from(value: DriveRole) -> Self {
        value as Self
    }
}

bitflags! {
    /// Status register
    ///
    /// See more informations on the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#Status_Register_.28I.2FO_base_.2B_7.29)
    #[derive(Debug, Clone, Copy)]
    struct Status: u8 {
        /// Indicates an error occurred. Send a new command to clear it (or nuke it with a Software Reset).
        const ERR   = 0b0000_0001;

        /// Index. Always set to zero.
        const IDX   = 0b0000_0010;

        /// Corrected data. Always set to zero.
        const CORR  = 0b0000_0100;

        /// Set when the drive has PIO data to transfer, or is ready to accept PIO data.
        const DRQ   = 0b0000_1000;

        /// Overlapped Mode Service Request.
        const SRV   = 0b0001_0000;

        /// Drive Fault Error (does not set ERR).
        const DF    = 0b0010_0000;

        /// Bit is clear when drive is spun down, or after an error. Set otherwise.
        const RDY   = 0b0100_0000;

        /// Indicates the drive is preparing to send/receive data (wait for it to clear). In case of 'hang' (it never clears), do a software reset.
        const BSY   = 0b1000_0000;
    }
}

bitflags! {
    /// Device control register
    ///
    /// See more informations on the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#Device_Control_Register_.28Control_base_.2B_0.29)
    struct DeviceControl: u8 {
        /// Set this to stop the current device from sending interrupts.
        const NIEN   = 0b0000_0010;

        /// Set, then clear (after 5us), this to do a "Software Reset" on all ATA drives on a bus, if one is misbehaving.
        const SRST  = 0b0000_0100;

        /// Set this to read back the High Order Byte of the last LBA48 value sent to an IO port.
        const HOB   = 0b1000_0000;
    }
}

impl From<u8> for Status {
    fn from(value: u8) -> Self {
        Self::from_bits_retain(value)
    }
}

/// Representation of an ATA bus
///
/// See more informations on the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#Registers)
#[derive(Debug)]
#[allow(unused)]
struct Bus {
    /// Read/write PIO data bytes
    data: Port<u16>,

    /// Used to retrieve any error generated by the last ATA command executed
    error: PortReadOnly<u8>,

    /// Used to control command specific interface features
    features: PortWriteOnly<u8>,

    /// Number of sectors to read/write (0 is a special value)
    sector_count: Port<u8>,

    /// This is CHS / LBA28 / LBA48 specific
    sector_number: Port<u8>,

    /// Partial Disk Sector address
    cylinder_low: Port<u8>,

    /// Partial Disk Sector address
    cylinder_high: Port<u8>,

    /// Used to select a drive and/or head. Supports extra address/flag bits
    drive_head: Port<u8>,

    /// Used to read the current status
    status: PortReadOnly<u8>,

    /// Used to send ATA commands to the device
    command: PortWriteOnly<u8>,

    /// A duplicate of the Status Register which does not affect interrupts
    alternate_status: PortReadOnly<u8>,

    /// Used to reset the bus or enable/disable interrupts
    device_control: PortWriteOnly<u8>,

    /// Provides drive select and head select information
    drive_address: PortReadOnly<u8>,
}

impl Bus {
    /// Returns a new bus from the data and control base address
    const fn new(data_bar: u16, control_bar: u16) -> Self {
        Self {
            data: Port::new(data_bar),
            error: PortReadOnly::new(data_bar + 1),
            features: PortWriteOnly::new(data_bar + 1),
            sector_count: Port::new(data_bar + 2),
            sector_number: Port::new(data_bar + 3),
            cylinder_low: Port::new(data_bar + 4),
            cylinder_high: Port::new(data_bar + 5),
            drive_head: Port::new(data_bar + 6),
            status: PortReadOnly::new(data_bar + 7),
            command: PortWriteOnly::new(data_bar + 7),
            alternate_status: PortReadOnly::new(control_bar + 2),
            device_control: PortWriteOnly::new(control_bar + 2),
            drive_address: PortReadOnly::new(control_bar + 3),
        }
    }

    /// Reads the `status` register
    ///
    /// The register is read 15 times before returning the value as suggested in the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#400ns_delays)
    fn status(&mut self) -> Status {
        for _ in 0..15_usize {
            // SAFETY: the status register is assumed to be well-defined
            unsafe {
                self.status.read();
            };
        }
        // SAFETY: the status register is assumed to be well-defined
        unsafe { self.status.read() }.into()
    }

    /// Waits until the bus is ready to read or write data
    ///
    /// This function is intended to be used after a command have been issued.
    ///
    /// Returns an error if the `status` port indicates an error
    fn wait_ready(&mut self) -> Result<()> {
        loop {
            let status = self.status();

            if status.intersects(Status::ERR | Status::DF) || (!status).intersects(Status::RDY) {
                return Err(anyhow::Error::msg("Drive Fault"));
            };

            if status.intersects(Status::BSY) {
                continue;
            };

            if status.intersects(Status::DRQ | Status::RDY) {
                return Ok(());
            }
        }
    }

    /// Sends a I/O command to the ATA bus
    ///
    /// # Safety
    ///
    /// Must ensure that `lba_start` and `sector` are well-defined
    #[allow(clippy::cast_possible_truncation)]
    unsafe fn initialize_io_pio(
        &mut self,
        role: DriveRole,
        lba_start: usize,
        sector_count: usize,
        command_24_bits: Command,
        command_48_bits: Command,
    ) {
        if lba_start <= MAX_LBA_28_VALUE {
            self.drive_head.write(0xE0 | Into::<u8>::into(role) | ((lba_start >> 24_u8) as u8 & 0x0F));
            self.sector_count.write(sector_count as u8);
            self.cylinder_high.write((lba_start >> 16_u8) as u8);
            self.cylinder_low.write((lba_start >> 8_u8) as u8);
            self.sector_number.write(lba_start as u8);
            self.command.write(command_24_bits.into());
        } else {
            self.drive_head.write(0x40 | Into::<u8>::into(role));
            self.sector_count.write((sector_count >> 8_u8) as u8);
            self.cylinder_high.write((lba_start >> 40_u8) as u8);
            self.cylinder_low.write((lba_start >> 32_u8) as u8);
            self.sector_number.write((lba_start >> 24_u8) as u8);
            self.sector_count.write(sector_count as u8);
            self.cylinder_high.write((lba_start >> 16_u8) as u8);
            self.cylinder_low.write((lba_start >> 8_u8) as u8);
            self.sector_number.write(lba_start as u8);
            self.command.write(command_48_bits.into());
        };
    }

    /// Sends a `Read PIO` command to the ATA bus and returns the number of blocks read by the function.
    ///
    /// This function does not perform any bound check.
    ///
    /// See the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#x86_Directions) for more informations on the algorithm
    ///
    /// # Safety
    ///
    /// Must ensure that `lba_start` and `sector` are well-defined
    unsafe fn read_pio(&mut self, buffer: &mut [u8], role: DriveRole, lba_start: usize, sector_count: usize) -> Result<usize> {
        if sector_count == 0 {
            return Ok(0);
        }

        self.wait_ready()?;

        self.initialize_io_pio(role, lba_start, sector_count, Command::ReadSectors, Command::ReadSectorsExt);

        let mut buffer_offset = 0;
        for _ in lba_start..(lba_start + sector_count) {
            self.wait_ready()?;

            for chunk in buffer.get_unchecked_mut(buffer_offset..(buffer_offset + BLOCK_SIZE)).chunks_exact(2) {
                let word = u16::from(*chunk.get_unchecked(1)) << 8_u8 | u16::from(*chunk.get_unchecked(0));
                self.data.write(word);
            }
            buffer_offset += BLOCK_SIZE;
        }

        self.wait_ready()?;
        Ok(sector_count)
    }

    /// Sends a `Write PIO` command to the ATA bus
    ///
    /// This function does not perform any bound check.
    ///
    /// See the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#x86_Directions) for more informations on the algorithm
    ///
    /// # Safety
    ///
    /// Must ensure that `lba_start` and `sector` are well-defined
    unsafe fn write_pio(&mut self, buffer: &mut [u8], role: DriveRole, lba_start: usize, sector_count: usize) -> Result<usize> {
        if sector_count == 0 {
            return Ok(0);
        }

        self.wait_ready()?;

        self.initialize_io_pio(role, lba_start, sector_count, Command::WriteSectors, Command::WriteSectorsExt);

        let mut buffer_offset = 0;
        for _ in lba_start..(lba_start + sector_count) {
            self.wait_ready()?;

            for chunk in buffer.get_unchecked_mut(buffer_offset..(buffer_offset + BLOCK_SIZE)).chunks_exact(2) {
                let word = u16::from(*chunk.get_unchecked(1)) << 8_u8 | u16::from(*chunk.get_unchecked(0));
                self.data.write(word);
            }
            buffer_offset += BLOCK_SIZE;
        }

        self.wait_ready()?;

        let cache_flush_cmd = if lba_start < MAX_LBA_28_VALUE { Command::FlushCache } else { Command::FlushCacheExt };
        self.command.write(cache_flush_cmd as u8);
        self.wait_ready()?;

        Ok(sector_count)
    }

    /// Returns an [`IdentifyDeviceData`], if it exists, from the bus on which the device is set and its role
    ///
    /// See the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#IDENTIFY_command) for more informations
    fn identify_data(&mut self, role: DriveRole) -> Result<IdentifyDeviceData> {
        self.wait_ready()?;

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.drive_head.write(0xA0 | u8::from(role));
        };

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.sector_count.write(0_u8);
        };

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.cylinder_high.write(0_u8);
        };

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.cylinder_low.write(0_u8);
        };

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.sector_number.write(0_u8);
        };

        // SAFETY: it is assume here that every register are well-defined
        unsafe {
            self.command.write(Command::IdentifyDevice.into());
        };

        if self.status().is_empty() {
            error!("Identify Device: tried to identify the {:?} ATA drive on the bus {:?} but it doesn't exist", role, self);
            return Err(anyhow::Error::msg("Identify Device: tried to identify an ATA drive that do not exist"));
        }

        while self.status().intersects(Status::BSY) {
            // SAFETY: it is assume here that every register are well-defined
            let lba_high = unsafe { self.cylinder_high.read() };

            // SAFETY: it is assume here that every register are well-defined
            let lba_mid = unsafe { self.cylinder_low.read() };

            if lba_high != 0 || lba_mid != 0 {
                return Err(anyhow::Error::msg("Identify Device: tried to identify an non-ATA drive"));
            }
        }

        // SAFETY: it is assume here that every register are well-defined
        let lba_high = unsafe { self.cylinder_high.read() };

        // SAFETY: it is assume here that every register are well-defined
        let lba_mid = unsafe { self.cylinder_low.read() };

        match DriveType::from_lba(lba_mid, lba_high) {
            Some(DriveType::Pata) => {
                let mut buffer = [0_u8; BLOCK_SIZE];
                self.wait_ready()?;
                for chunk in buffer.chunks_exact_mut(2) {
                    // SAFETY: it is assume here that every register are well-defined
                    let word = unsafe { self.data.read() };

                    // SAFETY: chunks are 2-byte longs
                    unsafe { chunk.get_unchecked_mut(0) }.clone_from(&((word & 0x00FF) as u8));

                    // SAFETY: chunks are 2-byte longs
                    unsafe { chunk.get_unchecked_mut(1) }.clone_from(&((word >> 8_u8) as u8));
                }
                self.wait_ready()?;
                Ok(IdentifyDeviceData::from(buffer))
            },
            Some(drive_type) => Err(anyhow::Error::msg(format!("Identify Device: unsupported {drive_type:?} type"))),
            None => Err(anyhow::Error::msg("Identify Device: tried to identify a drive with an unknown type")),
        }
    }

    /// Resets **both** ATA devices on the bus
    ///
    /// See the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#Resetting_a_drive_.2F_Software_Reset) for more informations
    fn software_reset(&mut self) {
        // SAFETY: it is assumed here that the `device_control` register is well defined
        unsafe {
            self.device_control.write(DeviceControl::SRST.bits());
        };

        for _ in 0..10_usize {
            self.status();
        }

        // SAFETY: it is assumed here that the `device_control` register is well defined
        unsafe {
            self.device_control.write(0_u8);
        };
    }
}

/// Existing kinds of ATA devices
///
/// See the [OSdev wiki](https://wiki.osdev.org/PCI_IDE_Controller#Parallel.2FSerial_ATA.2FATAPI) for more informations
#[derive(Debug)]
#[allow(unused)]
enum DriveType {
    /// Used for most modern hard drives
    Sata,

    /// Commonly used for hard drives
    Pata,

    /// Used for most modern optical drives
    Satapi,

    /// Commonly used for optical drives
    Patapi,
}

impl DriveType {
    /// Returns the type of ATA device from the LBA mid and LBA high ports values
    ///
    /// This function must be run after an identify device command has been sent and before the read of the response.
    ///
    /// See the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode#Detecting_device_types) for more informations
    const fn from_lba(lba_mid: u8, lba_high: u8) -> Option<Self> {
        match (lba_mid, lba_high) {
            (0x00, 0x00) => Some(Self::Pata),
            (0x14, 0xEB) => Some(Self::Patapi),
            (0x3C, 0xC3) => Some(Self::Sata),
            (0x69, 0x96) => Some(Self::Satapi),
            _ => None,
        }
    }
}

/// Description of an ATA drive obtained from an `identify` command
///
/// This structure is 256 `ushort` long, so 512 bytes.
///
/// See more informations on [Microsoft website](https://learn.microsoft.com/en-us/windows-hardware/drivers/ddi/ata/ns-ata-_identify_device_data#syntax)
#[repr(packed)]
#[derive(Debug, Clone)]
#[allow(unused)]
struct IdentifyDeviceData {
    /// Contains general configuration information
    general_configuration: u16,

    /// Indicates the number of cylinders on the device
    num_cylinders: u16,

    /// Specific configuration
    specific_configuration: u16,

    /// Number of logical heads on the device
    num_heads: u16,

    /// This member is no longer used
    retired_1: [u16; 2],

    /// Indicates the number of sectors per track
    num_sectors_per_track: u16,

    /// Contains the first ID of the device's vendor
    vendor_unique_1: [u16; 3],

    /// Contains the serial number of the device
    serial_number: [u8; 20],

    /// This member is no longer used
    retired_2: [u16; 2],

    /// This member is obsolete. Do not use.
    obsolete_1: u16,

    /// Contains the revision number of the device's firmware
    firmware_revision: [u8; 8],

    /// Contains the device's model number
    model_number: [u8; 40],

    /// Contains the maximum number of blocks allowed in a single transfer
    max_blocks_per_transfer: u8,

    /// Contains the second ID of the device's vendor
    vendor_unique_2: u8,

    /// Trusted computing
    trusted_computing: u16,

    /// Capabilities
    capabilities: u16,

    /// Reserved
    reserved_word_50: u16,

    /// This member is obsolete. Do not use.
    obsolete_words_51: [u16; 2],

    /// Contains a bitfield whose bits indicate which of the bytes in the identify data package contain valid address translation
    /// information.
    ///
    /// For more information about how this bitfield is defined, see the ATA/ATAPI specification.
    translation_fields_valid: u8,

    /// Free fall control sensitivity
    free_fall_control_sensitivity: u8,

    /// Indicates the number of cylinders on the device
    num_current_cylinders: u16,

    /// Indicates the number of heads on the device
    num_current_heads: u16,

    /// Indicates the number of sectors per track
    current_sectors_per_track: u16,

    /// Indicates the number of sectors on the device
    current_sector_capacity: u32,

    /// Indicates the multisector setting
    current_multi_sector_setting: u8,

    /// Byte 119
    ///
    /// Contains the following informations :
    /// - `MultiSectorSettingValid[1]`: Indicates when TRUE that the multisector setting is valid
    /// - `ReservedByte59[3]`: Reserved
    /// - `SanitizeFeatureSupported[1]`: Indicates whether the device supports sanitize
    /// - `CryptoScrambleExtCommandSupported[1]`: Indicates whether the device supports the crypto scramble command
    /// - `OverwriteExtCommandSupported[1]`: Indicates whether the device supports the overwrite command
    /// - `BlockEraseExtCommandSupported[1]`: Indicates whether the device supports block erase
    byte_119: u8,

    /// Indicates the total number of user-addressable sectors
    user_addressable_sectors: u32,

    /// This member is obsolete. Do not use.
    obsolete_word_62: u16,

    /// Indicates which DMA modes the device supports
    multiword_dma_support: u8,

    /// Indicates which DMA modes are currently selected
    multiword_dma_active: u8,

    /// Indicates PIO mode support
    advanced_pio_modes: u8,

    /// Reserved
    reserved_byte_64: u8,

    /// Indicates the minimum multiword DMA transfer cycle time per word
    minimum_mw_transfer_cycle_time: u16,

    /// Indicates the recommended multiword DMA transfer cycle time per word
    recommended_mw_transfer_cycle_time: u16,

    /// Indicates the minimum PIO transfer cycle time without flow control
    minimum_pio_cycle_time: u16,

    /// Indicates the minimum PIO transfer cycle time with IORDY flow control
    minimum_pio_cycle_time_io_rdy: u16,

    /// Additional supported
    additional_supported: u16,

    /// Reserved
    reserved_words_70: [u16; 5],

    /// Indicates the maximum queue depth
    queue_depth: u16,

    /// Serial ATA capabilities
    serial_ata_capabilities: u32,

    /// Serial ATA features supported
    serial_ata_features_supported: u16,

    /// Serial ATA features enabled
    serial_ata_features_enabled: u16,

    /// Indicates the device's major revision number
    major_revision: u16,

    /// Indicates the device's minor revision number
    minor_revision: u16,

    /// Command set support
    command_set_support: [u16; 3],

    /// Command set active
    command_set_active: [u16; 3],

    /// Contains a bitmap indicating which ultraDMA modes the device supports
    ultra_dma_support: u8,

    /// Contains a bitmap indicating which ultraDMA modes are selected
    ultra_dma_active: u8,

    /// Normal security erase unit
    normal_security_erase_unit: u16,

    /// Enhanced security erase unit
    enhanced_security_erase_unit: u16,

    /// Current APM level
    current_apm_level: u8,

    /// Reserved
    reserved_word_91: u8,

    /// Master password ID
    master_password_id: u16,

    /// Indicates the result of a hardware reset.
    ///
    /// For more information about the values assigned to this member, see the ATA/ATAP specification.
    hardware_reset_result: u16,

    /// Indicates the current acoustic management value
    current_acoustic_value: u8,

    /// Contain the device vendor's recommended acoustic management value
    recommended_acoustic_value: u8,

    /// Stream min request size
    stream_min_request_size: u16,

    /// Streaming transfer time DAM
    streaming_transfer_time_dma: u16,

    /// Streaming access latency DMA PIO
    streaming_access_latency_dma_pio: u16,

    /// Streaming perf granularity
    streaming_perf_granularity: u32,

    /// Contains the maximum user LBA for the 48-bit address feature set
    max_48_bit_lba: u64,

    /// Streaming transfer time
    streaming_transfer_time: u16,

    /// DSM cap
    dsm_cap: u16,

    /// Physical logical sector size
    physical_logical_sector_size: u16,

    /// Inter seek delay
    inter_seek_delay: u16,

    /// World wide name
    world_wide_name: [u16; 4],

    /// Reserved for world wide name 128
    reserved_for_world_wide_name_128: [u16; 4],

    /// Reserved for tlc technical report
    reserved_for_tlc_technical_report: u16,

    /// Words per logical sector
    words_per_logical_sector: u32,

    /// Command set support ext
    command_set_support_ext: u16,

    /// Command set active ext
    command_set_active_ext: u16,

    /// Reserved for expanded support and active
    reserved_for_expanded_support_and_active: [u16; 6],

    /// Indicates when TRUE that the device supports media status notification
    msn_support: u16,

    /// Contains a bitmap that indicates the security status.
    ///
    /// For more information about the meaning of each individual bit, see the ATA/ATAPI specification
    security_status: u16,

    /// Reserved
    reserved_word_129: [u16; 31],

    /// CFA power model
    cfa_power_mode1: u16,

    /// Words 161-168
    reserved_for_cfa_word_161: [u16; 7],

    /// Nominal form factor
    nominal_form_factor: u16,

    /// Data set management feature
    data_set_management_feature: u16,

    /// Additional product ID
    additional_product_id: [u16; 4],

    /// Reserved for CFA word 174
    reserved_for_cfa_word_174: [u16; 2],

    /// Current media serial number
    current_media_serial_number: [u16; 30],

    /// SCT command transport
    sct_command_transport: u16,

    /// Words 207-208
    reserved_word_207: [u16; 2],

    /// Block alignment
    block_alignment: u16,

    /// Words 210-211
    write_read_verify_sector_count_mode_3_only: [u16; 2],

    /// Words 212-213
    write_read_verify_sector_count_mode_2_only: [u16; 2],

    /// NV cache capabilities
    nv_cache_capabilities: u16,

    /// NV cache size LSW
    nv_cache_size_lsw: u16,

    /// NV cache size MSW
    nv_cache_size_msw: u16,

    /// Nominal media rotation rate
    nominal_media_rotation_rate: u16,

    /// Reserved word 218
    reserved_word_218: u16,

    /// NV cache time to spin up in seconds
    nv_cache_time_to_spin_up_in_seconds: u8,

    /// NV cache option reserved
    nv_cache_option_reserved: u8,

    /// Write read verify sector count mode
    write_read_verify_sector_count_mode: u8,

    /// Reserved word 220
    reserved_word_220: u8,

    /// Reserved word 220
    reserved_word_221: u16,

    /// Transport major version
    transport_major_version: u16,

    /// Transport minor version
    transport_minor_version: u16,

    /// Reserved word 224
    reserved_word_224: [u16; 6],

    /// Extended num of user addressable sectors
    extended_num_of_user_addressable_sectors: u64,

    /// Min blocks per download microcode
    min_blocks_per_download_microcode: u16,

    /// Max blocks per download microcode
    max_blocks_per_download_microcode: u16,

    /// Reserved word 236
    reserved_word_236: [u16; 19],

    /// Indicates the disk signature
    signature: u8,

    /// Indicates the checksum
    checksum: u8,
}

impl From<[u8; IDENTIFY_DATA_INFORMATIONS_SIZE]> for IdentifyDeviceData {
    fn from(value: [u8; IDENTIFY_DATA_INFORMATIONS_SIZE]) -> Self {
        // SAFETY: it is assumed that the given raw bytes points to an `IdentifyDeviceData` structure
        unsafe { mem::transmute(value) }
    }
}

/// Common interface for ATA drives
///
/// More informations on ATA drives on the [OSdev wiki](https://wiki.osdev.org/ATA_PIO_Mode)
#[derive(Debug, Clone)]
pub struct Drive {
    /// Reference to the bus to which the drive is connected
    ///
    /// This bus is shared with the other [`Drive`] connected to this bus.
    bus: Arc<Mutex<Bus>>,

    /// Kind of the ATA drive (master/slave)
    role: DriveRole,

    /// Data retrieved after a `identify` command
    identify_data: IdentifyDeviceData,
}

impl Manipulation for Drive {
    fn block_size(&self) -> usize {
        BLOCK_SIZE
    }

    fn len(&self) -> usize {
        self.block_size() * self.size_in_blocks()
    }

    fn read(&mut self, buffer: &mut [u8], offset: usize, length: usize) -> Result<usize> {
        self.read_pio(buffer, offset, length)
    }

    fn write(&mut self, buffer: &mut [u8], offset: usize, length: usize) -> Result<usize> {
        self.write_pio(buffer, offset, length)
    }

    fn flush(&mut self) -> anyhow::Result<()> {
        Ok(())
    }
}

impl Device for Drive {
    #[allow(clippy::cast_possible_truncation)]
    fn size_in_blocks(&self) -> usize {
        if self.identify_data.user_addressable_sectors == 0 {
            self.identify_data.max_48_bit_lba as usize
        } else {
            self.identify_data.user_addressable_sectors as usize
        }
    }
}

impl Drive {
    /// Initializes, if present, the ATA device on the given `bus` at the given `role`
    ///
    /// This function should be only called once for each tuple `(bus, role)`
    fn initialize(bus_ref: Arc<Mutex<Bus>>, role: DriveRole) -> Result<Self> {
        bus_ref.lock().software_reset();
        let identify_data = bus_ref.lock().identify_data(role)?;

        if identify_data.capabilities & 0b0000_0010_0000_0000 == 0 {
            Err(anyhow::Error::msg("Initialize Drive: the found ATA drive doesn't support LBA addressing mode"))
        } else {
            Ok(Self {
                bus: bus_ref,
                role,
                identify_data,
            })
        }
    }

    /// Reads `length` blocks of data from the drive starting at the given `offset` (in **blocks**) and returns the result in the
    /// `buffer`
    ///
    /// Returns the number of **blocks** read (and not bytes)
    ///
    /// # Safety
    ///
    /// Must ensure that the given buffer has a byte length of at least `length * BLOCK_SIZE`
    fn read_pio(&mut self, buffer: &mut [u8], offset: usize, length: usize) -> Result<usize> {
        if offset > self.size_in_blocks() {
            error!("Read PIO: offset out of bounds");
            return Err(anyhow::Error::msg("Read PIO: Offset out of bounds"));
        }

        trace!("Read {} blocks at {} from an ATA drive", length, offset);

        if length > self.identify_data.max_blocks_per_transfer.into() {
            error!(
                "Read PIO: cannot read {} blocks with a max_block_per_transfer of {}",
                length, self.identify_data.max_blocks_per_transfer
            );
            return Err(anyhow::Error::msg("Read PIO: cannot read more blocks that the drive's max"));
        }

        // SAFETY: if is assumed that the drive's bus and informations are well defined
        unsafe { self.bus.lock().read_pio(buffer, self.role, offset, length) }
    }

    /// Writes `length` blocks of data from the given `buffer` to the drive starting at the given `offset` (in **blocks**)
    ///
    /// Returns the number of **blocks** read (and not bytes)
    ///
    /// # Safety
    ///
    /// Must ensure that the given buffer has a byte length of at least `length * BLOCK_SIZE`
    fn write_pio(&mut self, buffer: &mut [u8], offset: usize, length: usize) -> Result<usize> {
        if offset > self.size_in_blocks() {
            error!("Write PIO: offset out of bounds");
            return Err(anyhow::Error::msg("Write PIO: Offset out of bounds"));
        }

        trace!("Write {} blocks at {} from an ATA drive", length, offset);

        if length > self.identify_data.max_blocks_per_transfer.into() {
            error!(
                "Write PIO: cannot read {} blocks with a max_block_per_transfer of {}",
                length, self.identify_data.max_blocks_per_transfer
            );
            return Err(anyhow::Error::msg("Write PIO: cannot read more blocks that the drive's max"));
        }

        // SAFETY: if is assumed that the drive's bus and informations are well defined
        unsafe { self.bus.lock().write_pio(buffer, self.role, offset, length) }
    }
}

/// Interface for the electrical specification of the cables connecting ATA drives
///
/// See the [OSdev wiki](https://wiki.osdev.org/PCI_IDE_Controller#IDE_Interface) for more informations
#[derive(Debug)]
pub struct IdeController {
    /// Primary master drive
    primary_master: Option<Arc<Mutex<Drive>>>,

    /// Primary slave drive
    primary_slave: Option<Arc<Mutex<Drive>>>,

    /// Secondary master drive
    secondary_master: Option<Arc<Mutex<Drive>>>,

    /// Secondary slave drive
    secondary_slave: Option<Arc<Mutex<Drive>>>,
}

impl Controller for IdeController {
    fn devices<'device>(&self) -> vec::Vec<Arc<Mutex<dyn Device + Send + 'device>>> {
        vec![self.primary_master.clone(), self.primary_slave.clone(), self.secondary_master.clone(), self.secondary_slave.clone()]
            .iter()
            .filter_map(|device_option| {
                device_option
                    .as_ref()
                    .map_or_else(|| None, |device| Some(Arc::clone(device) as Arc<Mutex<dyn Device + Send + 'device>>))
            })
            .collect()
    }
}

impl IdeController {
    /// Creates a new IDE controller from a given PCI device
    pub fn new(device: &pci::Device) -> Option<Self> {
        let Header::X0 {
            bar0,
            bar1,
            bar2,
            bar3,
            ..
        } = device.header else {
            error!("PCI devices with an header type different than 0x0 are not supported");
            return None;
        };

        let primary_bus_port = match bar0 {
            0x0000 | 0x0001 => DEFAULT_PRIMARY_BUS_REGISTER,
            other => {
                warn!("Unusual ATA bar0 address : {:x}", bar0);
                (other & 0x000_0FFFF) as u16
            },
        };

        let primary_control_status_port = match bar1 {
            0x0000 | 0x0001 => DEFAULT_PRIMARY_CONTROL_STATUS_REGISTER,
            other => {
                warn!("Unusual ATA bar1 address : {:x}", bar1);
                (other & 0x0000_FFFF) as u16
            },
        };

        let secondary_bus_port = match bar2 {
            0x0000 | 0x0001 => DEFAULT_SECONDARY_BUS_REGISTER,
            other => {
                warn!("Unusual ATA bar2 address : {:x}", bar2);
                (other & 0x0000_FFFF) as u16
            },
        };

        let secondary_control_status_port = match bar3 {
            0x0000 | 0x0001 => DEFAULT_SECONDARY_CONTROL_STATUS_REGISTER,
            other => {
                warn!("Unusual ATA bar3 address : {:x}", bar3);
                (other & 0x0000_FFFF) as u16
            },
        };

        let primary_bus = Arc::new(Mutex::new(Bus::new(primary_bus_port, primary_control_status_port)));
        let secondary_bus = Arc::new(Mutex::new(Bus::new(secondary_bus_port, secondary_control_status_port)));

        let primary_master = Drive::initialize(Arc::clone(&primary_bus), DriveRole::Master);
        let primary_slave = Drive::initialize(Arc::clone(&primary_bus), DriveRole::Slave);
        let secondary_master = Drive::initialize(Arc::clone(&secondary_bus), DriveRole::Master);
        let secondary_slave = Drive::initialize(Arc::clone(&secondary_bus), DriveRole::Slave);

        info!(
            "Found an ATA controller at {:?} containing :\n\t* Primary master: {:?}\n\t* Primary slave: {:?}\n\t* Secondary master: {:?}\n\t* Secondary slave: {:?}",
            device.location, primary_master, primary_slave, secondary_master, secondary_slave
        );

        let to_ref = |drive| Arc::new(Mutex::new(drive));

        Some(Self {
            primary_master: primary_master.ok().map(to_ref),
            primary_slave: primary_slave.ok().map(to_ref),
            secondary_master: secondary_master.ok().map(to_ref),
            secondary_slave: secondary_slave.ok().map(to_ref),
        })
    }
}
