//! Device interface

use alloc::sync::Arc;
use alloc::vec::Vec;

use conquer_once::spin::OnceCell;
use log::info;
use spin::Mutex;

use crate::kernel::device::ata::IdeController;

#[allow(clippy::same_name_method)]
mod ata;
mod block;
mod pci;
pub mod storage;

/// List of references towards all initialized storage controllers
pub static STORAGE_CONTROLLERS: OnceCell<Vec<Arc<Mutex<dyn storage::Controller + Send>>>> = OnceCell::uninit();

/// Initializes the devices
///
/// Currently, is is only used for storage devices.
pub fn init() {
    let devices = pci::scan();
    info!("{} devices found during PCI scan", devices.len());

    let storage_controllers = STORAGE_CONTROLLERS.get_or_init(|| {
        devices
            .values()
            .flatten()
            .filter_map(|device| {
                Some(Arc::new(Mutex::new(IdeController::new(device)?)) as Arc<Mutex<dyn storage::Controller + Send>>)
            })
            .collect::<Vec<Arc<Mutex<dyn storage::Controller + Send>>>>()
    });

    info!(
        "{} storage controller(s) and {} storage drive(s) found",
        storage_controllers.len(),
        storage_controllers
            .iter()
            .fold(0_usize, |acc, controller| acc + controller.lock().devices().len())
    );
}
